#include <sys/poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/joystick.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

//#define UFDS_COUNT 10
#define UFDS_COUNT 1

void select_call(const char *device)
{
	int result, fd;
	fd_set tmpfds, readfds, writefds, exceptfds;

	fd = open(device, O_RDONLY | O_NONBLOCK);
	if(fd == -1)
	{
		fprintf(stderr, "Error opening %s: %s\n",
			device, strerror(errno));
		exit(1);
	}

	FD_ZERO(&tmpfds);
	FD_SET(fd, &tmpfds);
	for(;;)
	{
		writefds = exceptfds = readfds = tmpfds;
		result = select(fd+1, &readfds, &writefds, &exceptfds, NULL);
		if(result == -1)
		{
			perror("Select error");
			exit(1);
		}
		if(FD_ISSET(fd, &readfds))
		{
			struct js_event event;
			int size;
			printf("%s readable\n", __FUNCTION__);
			size = read(fd, &event, sizeof(event));
			if(size == -1)
			{
				fprintf(stderr, "%s Error reading data\n",
					__FUNCTION__);
				if(errno == ENODEV)
					exit(1);
			}
		}
		if(FD_ISSET(fd, &writefds))
		{
			printf("%s writable\n", __FUNCTION__);
		}
		if(FD_ISSET(fd, &exceptfds))
		{
			printf("%s exception\n", __FUNCTION__);
			break;
		}
	}
}

void poll_call(const char *device)
{
	int i, result, fd;
	struct pollfd ufds[UFDS_COUNT];
	fd = open(device, O_RDONLY | O_NONBLOCK);
	if(fd == -1)
	{
		fprintf(stderr, "Error opening %s: %s\n",
			device, strerror(errno));
		exit(1);
	}
	for(i=0; i<UFDS_COUNT; i++)
	{
		ufds[i].fd = fd;
		ufds[i].events = POLLIN | POLLPRI | POLLOUT | POLLERR |
			POLLHUP | POLLNVAL;
		ufds[i].revents = 0;
	}
	for(;;)
	{
		result = poll(ufds, UFDS_COUNT, -1);
		if(result == -1)
		{
			perror("poll error");
			exit(1);
		}
		for(i=0; i<result; i++)
		{
			printf("poll %d, ", i);
			if(ufds[i].revents & POLLIN ||
				ufds[i].revents & POLLERR)
			{
				struct js_event event;
				int size;
				if(ufds[i].revents & POLLIN)
					printf("POLLIN There is data to read");
				else if(ufds[i].revents & POLLERR)
					printf("POLLERR Error condition");
				size = read(fd, &event, sizeof(event));
				if(size == -1)
				{
					fprintf(stderr,
						"%s Error reading data\n",
						__FUNCTION__);
					if(errno == ENODEV)
						exit(1);
				}
			}
			if(ufds[i].revents & POLLPRI)
				printf("POLLPRI There is urgent data to read");
			if(ufds[i].revents & POLLOUT)
				printf("POLLOUT Writing now will not block");
			if(ufds[i].revents & POLLHUP)
				printf("POLLHUP Hung up");
			if(ufds[i].revents & POLLNVAL)
				printf("POLLNVAL Invalid request: fd not open");
			printf("\n");
		}
	}
}

void read_call(const char *device)
{
	struct js_event event;
	int size, fd;
	/* Don't block here. */
	fd = open(device, O_RDONLY);
	if(fd == -1)
	{
		fprintf(stderr, "Error opening %s: %s\n",
			device, strerror(errno));
		exit(1);
	}
	for(;;)
	{
		size = read(fd, &event, sizeof(event));
		if(size == -1)
		{
			fprintf(stderr, "%s Error reading data: %s\n",
				__FUNCTION__, strerror(errno));
			exit(1);
		}
		else
		{
			printf("%s read %d bytes\n", __FUNCTION__, size);
		}
	}
}

int main(int argc, char **argv)
{
	int result;
	const char *device;
	if(argc != 2)
	{
		fprintf(stderr, "Usage: %s /dev/jsN\n", argv[0]);
		return -1;
	}
	device = argv[1];

	result = fork();
	if(result==-1)
	{
		perror("Error on fork.");
		return -1;
	}
	if(result)
	{
		result = fork();
		if(result == -1)
		{
			perror("Error on fork.");
			return -1;
		}
		if(result)
		{
			select_call(device);
		}
		else
		{
			read_call(device);
		}
	}
	else
	{
		poll_call(device);
	}
	return 0;
}
